#SkIndigo Ruby code is copyright Dale Martens 2009, all rights reserved, and
#Glare Technologies Limited, 2014 - 
##################################Mesh Builder#############################



# An instance, these will be exported as instances in the scene file, e.g. as <model2> elements.
# An instance holds a reference to either a Sketchup::ComponentDefinition or Sketchup::Group which it represents,
# or both of those may be nil in which case the instance represents the top-level geometry (faces) in the scene.
class IndigoInstance

	attr_accessor :material, :transformation, :definition, :su_instance, :group

	def initialize
		@material = nil  # Sketchup::Material, may be nil
		@transformation = nil # Sketchup::Transformation
		@definition = nil # Sketchup::ComponentDefinition, may be nil
		@group = nil # Sketchup::Group, may be nil
		@su_instance = nil # Sketchup::ComponentInstance, may be null
	end
	
	# Will not necessarily be globally-unique.
	def name
		return @definition.name if @definition
		return @group.name if @group
		
		return "Object"
	end
end


class IndigoMeshBuilder

	# In SketchUp 2014 this class was changed from Set to Sketchup::Set
	# in order to avoid conflict with the Ruby Standard Library. 
	# from: http://www.sketchup.com/intl/en/developer/docs/ourdoc/set
	if defined?(Sketchup::Set)
		Set = Sketchup::Set
	end

	# These class variables keep track of all the component meshes, instances, and spheres in the model
	# they are in effect global so that when an animation is exported, the meshes and instances are already saved in memory
	# so they don't have to be 'collected' again by traversing all the entites in the model
	@@components_to_export = Set.new  # 'used' component definitions (in model and visible)
	
	# Sketchup::Entities from the top-level of the scene. type Sketchup::Entities
	@@top_level_entities_to_export = nil
	
	#  A list of instances created during process_entity_list, to be exported later.
	@@instances = []
	
	@@indigo_meshes = {}  # Constructed Indigo meshes, hash from Sketchup::ComponentDefinition to IndigoMesh
	@@top_level_entity_mesh = nil # IndigoMesh constructed from top-level geometry (faces).
	
	# Since a component definition with a given GUID will never change (if it does the GUID will change), keep a mesh cached for each GUID.
	@@meshes_for_comp_def_guid = {} # Hash from GUID string to IndigoMesh
	
	@@spheres = []
	@@images = []
	@@section_planes = []
	@@dict = "indigo07_materials"
	

	def initialize
		@scale = 0.0254 * Indigo.scale
		@num_new_meshes_exported = 0
	end

	def num_meshes()
		return @@indigo_meshes.keys.length + (@@top_level_entity_mesh ? 1 : 0)
	end
	
	def num_new_meshes_exported
		return @num_new_meshes_exported
	end

	# Meshes need to be cleared between different scenes, as the mesh holds a ComponentDefinition reference that will be deleted when a scene is closed.
	def IndigoMeshBuilder.clear_cached_meshes
		@@meshes_for_comp_def_guid = {}
	end
	
	
	# Reset the class variables
	def IndigoMeshBuilder.reset
		@@components_to_export = Set.new
		@@top_level_entities_to_export = nil
		
		@@instances = []
		
		@@indigo_meshes = {}
		@@top_level_entity_mesh = nil
		
		@@spheres=[]
		@@images=[]
		@@section_planes=[]
	end
	
	# Get the IndigoMesh for a particular IndigoInstance
	def get_indigomesh(i_instance)
	
		# An instance with nil ComponentDefinition ref and nil Group ref is referring to the top-level geometry.
		if !i_instance.definition && !i_instance.group
			return @@top_level_entity_mesh
		end
	
		if i_instance.definition
			return @@indigo_meshes[i_instance.definition]
		else
			return nil
		end
	end

	def reset_instances
		@@instances = []
		@@spheres=[]
		@@images=[]
	end


	def spheres
		return @@spheres
	end
	
	def instances
		return @@instances
	end

	def all_meshes_to_export
		return @@indigo_meshes.values + (@@top_level_entity_mesh ? [@@top_level_entity_mesh] : [])
	end

	def images
		return @@images
	end

	def section_planes
		return @@section_planes
	end

	def load_external_skp_defs()
		defns=Sketchup.active_model.definitions
		for defn in defns
			if IndigoMesh.external_skp?(defn)
				instances=defn.instances
				imesh=IndigoMesh.new(defn)
				external_def=defns.load(imesh.external_mesh_path)
				instances.each {|inst| inst.definition=external_def}
			end
		end
	end

	
	# defn will be a ComponentDefinition or Group
	def build_mesh_full_or_mats_only(defn, use_mesh_dir)
		#puts "build_mesh_full_or_mats_only(), defn.name: #{defn.name}, defn.guid: #{defn.guid}"
		
		mesh_path = File.join(use_mesh_dir, defn.guid + ".igmesh")
		
		if FileTest.exist?(mesh_path)
			# Mesh is already exported to disk.
			# So we don't need to completely build it.
			# All we need to do is get a list of the materials used by it.
			#puts "Mesh #{mesh_path} already exists on disk."
			
			return build_mesh_materials_only(defn, defn.entities)
		else
			#puts "Mesh #{mesh_path} does not already exist on disk."
			
			@num_new_meshes_exported += 1
		
			# If we have got this far, the mesh must be built or rebuilt
			return build_mesh(defn, defn.entities)
		end
	end

	# Get the path to the directory to save meshes (.igmesh files) to.
	# Will create the directory if it does not currently exist.
	def mesh_dir
		dir = File.join(SkIndigo.get_user_app_support_path, "skindigo_meshes")
		Dir.mkdir(dir) if !FileTest.exist?(dir)
		return dir
	end
	
	# Process a .mtl file (which will be referened by an .obj file)
	# Returns hash material name to hash of attributes, e.g. {"red_plastic" => {:r=>0.5, :g=>0.3, :b=>0.2}}
	# Note: the parsing technique below is slightly ghetto (using split() etc..), but Ruby doesn't make it easy to do parsing properly so this will do for now.
	def process_mtl_file(path)
		begin
			mtllib_file = File.new(path, "r")
			data = mtllib_file.read
			mtllib_file.close

			tokens = data.split

			result = {}
			
			i = 0
			while(i<tokens.size)
				
				# Advance to next 'newmtl' token
				while(i < tokens.size && tokens[i] != "newmtl")
					i += 1
				end
				if(i >= tokens.size)
					break
				end

				i += 1 # Advance past 'newmtl'
				mtlname = tokens[i]
				#puts "mtlname: '#{mtlname}'"
				i += 1 # Advance past mat name
								
				# Parse mat attributes
				mat_attributes = {}
				while(i < tokens.size)
					if tokens[i] == "Ka"
						i += 4 # Advance past "Ka" and 3 colour components afterwards
					elsif tokens[i] == "Kd"
						r = tokens[i + 1].to_f
						g = tokens[i + 2].to_f
						b = tokens[i + 3].to_f
						#puts "rgb: #{r}, #{g}, #{b}"
						mat_attributes[:r] = r
						mat_attributes[:g] = g
						mat_attributes[:b] = b
						i += 4
					elsif tokens[i] == "Ks"
						i += 4
					elsif tokens[i] == "map_Kd"
						mat_attributes[:map_Kd] = tokens[i + 1]
						i += 2
					elsif tokens[i] == "newmtl"
						break
					else
						# Unknown token
						#puts "Unknown token #{tokens[i]}"
						i += 1
					end
				end
				
				result[mtlname] = mat_attributes
			end
			return result
		rescue
			return {}
		end
	end
	
	# If the mesh is an external OBJ mesh, we need to read the OBJ file to get the material names from it.
	def build_mesh_materials_for_obj(defn)
		
		#puts "build_mesh_materials_for_obj()"
		
		im = IndigoMesh.new(defn)
		
		# Load the obj file and scan for used material names.
		file = File.new(im.external_mesh_path, "r")
		data = file.read
		file.close
		
		# Try and load the material file (.mtl) if present.
		# First, parse the path:
		mtllib_path = data.scan(/^mtllib (.*)$/)
		
		# puts "mtllib_path: #{mtllib_path.inspect}"
		
		material_attributes = {}
		if !mtllib_path.empty? && !mtllib_path[0].empty?
			path = mtllib_path[0][0]
				
			fullpath = File.dirname(im.external_mesh_path) + "/" + path
			
			material_attributes = process_mtl_file(fullpath)
			# puts "material_attributes: #{material_attributes.inspect}"
		end
		
		
		# ^:	Start of line
		# $:	End of line
		# (.*):	capture group
		used_mtls = data.scan(/^usemtl (.*)$/)
		
		used_mtls.each do |match|
			mat_name = match[0]
			
			# Find the material for the material name
			mat = Sketchup.active_model.materials[mat_name]
			
			if mat
				im.add_material(mat)
			else
				puts "Adding new material '#{mat_name}'..."
				
				# Add a new material!
				newmat = Sketchup.active_model.materials.add(mat_name)
				
				# Set some properties of the material based on what we parsed from the .mtl file.
				mat_attributes = material_attributes[mat_name]
				if mat_attributes
					if mat_attributes[:r] && mat_attributes[:g] && mat_attributes[:b]
						newmat.color = [mat_attributes[:r], mat_attributes[:g], mat_attributes[:b]]
					end
					if mat_attributes[:map_Kd]
						newmat.texture = File.join(File.dirname(im.external_mesh_path), mat_attributes[:map_Kd])
					end
				end
				
				im.add_material(newmat)
			end
		end
		
		# The mesh definition should have at least one material in its material list, even if it is nil.
		# Since it's nil it will get replaced with the scene default material or the component instance material.
		if used_mtls.empty?
			im.add_material(nil)
		end
		
		return im
	end
	
	
	# Iterates over all components, groups, and entity lists that we collected during process_entity_list(), and builds IndigoMesh objects for them.
	def build_meshes()
			
		#puts "IndigoMeshBuilder::build_meshes()"
		
		use_mesh_dir = mesh_dir
		
		#puts "----Exporting components----"
		
		num_components_to_export = @@components_to_export.length + 1
		i = 0
		
		@@components_to_export.each do |defn|
			# puts "Building mesh for component #{defn.name}"
			
			# See if mesh is already built and in memory.
			imesh = @@meshes_for_comp_def_guid[defn.guid]
			
			# Check to see if the mesh is still valid.  Sometimes a mesh can be deleted, but we still have a reference to it for some reason.
			# So do this check to catch/avoid the 'reference to deleted Entity' problem.
			if imesh && !imesh.is_valid()
				# puts "imesh was not valid!"
				imesh = nil
			end
			
			if imesh
				#puts "Mesh for component defn with guid #{defn.guid} already built (and was stored in  @@meshes_for_comp_def_guid)"
			else
				#puts "Building Mesh for component defn with guid #{defn.guid}"
				
				if IndigoMesh.external_obj?(defn)
					imesh = build_mesh_materials_for_obj(defn)
				else
					imesh = build_mesh_full_or_mats_only(defn, use_mesh_dir)
				end
				
				@@meshes_for_comp_def_guid[defn.guid] = imesh
			end

			@@indigo_meshes[defn] = imesh
			
			
			# Updated status text, only do every N meshes after a while.
			Sketchup.set_status_text("Building Meshes (#{i} / #{num_components_to_export})...") if (i <= 20) || (i % 10 == 0)
			i += 1
		end

		
		# Build mesh for top_level_entities_to_export
		if @@top_level_entities_to_export
			#puts "Building mesh for top_level_entities_to_export..."
			
			@@top_level_entity_mesh = build_mesh(nil, @@top_level_entities_to_export)
			@num_new_meshes_exported += 1
		end
	end

	
	def build_mesh_materials_only(defn, ents)
		
		#puts "build_mesh_materials_only()"
		
		im = IndigoMesh.new(defn)
		
		# This loop has to give the materials in the same order as build_mesh() does.

		ents.grep(Sketchup::Face) do |e|
			if e.visible? and e.layer.visible?
				mat = e.material
				back_mat = e.back_material
				if (mat==nil) and (back_mat!=nil)
					mat = back_mat
				end
		
				im.add_material(mat)
			end
		end
		
		return im
	end
	
	
	# image has type Skindio::Image
	# NOTE: Code similar to IndigoExporter::export_image() 
	def add_image_to_mesh(su_image, indigo_mesh, transform, mat)
		#puts "add_image_to_mesh()"
		position_data = []
		normal_data = []
		uv_data = []
		tri_data = []
		v_index = indigo_mesh.vert_pos_data.length / 3
	
		id = su_image.entityID.to_s
		path = su_image.path
		
		normal = su_image.normal
		origin = su_image.origin
		zrot = su_image.zrotation
		
		local_trans = Geom::Transformation.new(origin, normal)
		rot_trans = Geom::Transformation.rotation([0,0,0], Z_AXIS, zrot)
		transform = transform * (local_trans * rot_trans)

		transform = Geom::Transformation.scaling(@scale) * transform
		
		# NOTE: su_image.width is in inches.
		width_in = su_image.width.to_f
		height_in = su_image.height.to_f
		
		# Insert the sketchup image as the material.  This will be detected and exported as a proper material later, in IndigoExporter.export_scatter().
		mat_index = indigo_mesh.add_material(su_image)
		
		points = [
			transform * Geom::Point3d.new(0, 0, 0),
			transform * Geom::Point3d.new(width_in, 0, 0),
			transform * Geom::Point3d.new(width_in, height_in, 0),
			transform * Geom::Point3d.new(0, height_in, 0)
		]
		
		uvs = [
			Geom::Point3d.new(0, 0, 0),
			Geom::Point3d.new(1, 0, 0),
			Geom::Point3d.new(1, 1, 0),
			Geom::Point3d.new(0, 1, 0)
		]
		
		normals = [
			normal,
			normal,
			normal,
			normal
		]
		
		points.each do |point| 
			position_data << point.x
			position_data << point.y
			position_data << point.z
		end
		
		normals.each do |normal| 
			normal_data << normal.x
			normal_data << normal.y
			normal_data << normal.z
		end
		
		uvs.each do |uv|
			uv_data << uv.x
			uv_data << uv.y
		end
		
		v0 = v_index
		v1 = v_index + 1
		v2 = v_index + 2
		v3 = v_index + 3
		
		# Write tris for the image quad.
		# TODO: write as a quad?
		tri_data << v0 # vert indices
		tri_data << v1
		tri_data << v2
		tri_data << v0 # uv indices
		tri_data << v1
		tri_data << v2
		tri_data << mat_index
		
		tri_data << v0 # vert indices
		tri_data << v2
		tri_data << v3
		tri_data << v0 # uv indices
		tri_data << v2
		tri_data << v3
		tri_data << mat_index
	
		indigo_mesh.vert_pos_data = 	indigo_mesh.vert_pos_data.concat(position_data)
		indigo_mesh.vert_normal_data = 	indigo_mesh.vert_normal_data.concat(normal_data)
		indigo_mesh.uv_data = 			indigo_mesh.uv_data.concat(uv_data)
		indigo_mesh.triangle_data = 	indigo_mesh.triangle_data.concat(tri_data)	
	end
	
	
	
	def add_faces_to_mesh(ents, indigo_mesh, transform, instance_mat)
		#puts "add_faces_to_mesh()"
		position_data = []
		normal_data = []
		uv_data = []
		tri_data = []
		v_index = indigo_mesh.vert_pos_data.length / 3
		#puts "v_index: #{v_index}"
		num_faces = 0
		
		
		scale_factor = @scale
		
		ents.grep(Sketchup::Face) do |e|
			if e.visible? and e.layer.visible?
			
				face = e
				mat = face.material
				back_mat = face.back_material
				backface = false
				if (mat==nil) and (back_mat!=nil)
					backface = true
					mat = back_mat
				end
				
				mat = instance_mat if !mat # NEW
		
				
				mat_index = indigo_mesh.add_material(mat)
				#add_material_time += Time.now - t
				
				#t = Time.now
				if backface
					facemesh = face.mesh(7)
				else
					facemesh = face.mesh(5)
				end
				#make_face_mesh_time += Time.now - t
				

				frontface = !backface

				
				# Get vertex positions
				#t = Time.now
				facemesh_points = facemesh.points
				facemesh_points.each do |point| 
					#puts "point: #{point}"
					point = transform * point
					#puts "transformed point: #{point}"
					position_data << point.x * scale_factor
					position_data << point.y * scale_factor
					position_data << point.z * scale_factor
				end
				#get_vert_pos_time += Time.now - t
				#puts "Getting vertex positions took #{Time.now - t} s"
				
				# Get vertex normals
				#t = Time.now
				num_points = facemesh_points.length
				if frontface
					for i in 1..num_points
						n = facemesh.normal_at(i)
						n = transform * n
						normal_data << n.x
						normal_data << n.y
						normal_data << n.z
					end
				else
					for i in 1..num_points
						n = facemesh.normal_at(i)
						n = transform * n
						normal_data << -n.x
						normal_data << -n.y
						normal_data << -n.z
					end
				end
				#get_vert_normal_time += Time.now - t
				#puts "Getting vertex normals took #{Time.now - t} s"
				
				# Get vertex UVs
				#t = Time.now
				for i in 1..num_points
					uv = facemesh.uv_at(i, frontface)  # NOTE: can maybe use .uvs instead of .uv_at?
					#puts "uv: #{uv}"
					uv_data << uv.x  # TODO: divide by uv.z here?
					uv_data << uv.y
				end
				#get_uv_time += Time.now - t
				#puts "Getting vertex UVs took #{Time.now - t} s"
				
				
				# Get triangles
				#t = Time.now
				polygons = facemesh.polygons
				for poly in polygons
					v1 = poly[0]
					v2 = poly[1]
					v3 = poly[2]
					# Convert to positive values, then convert to zero-based indices.
					v1 = (v1 >= 0 ? v1 : -v1) + v_index - 1
					v2 = (v2 >= 0 ? v2 : -v2) + v_index - 1
					v3 = (v3 >= 0 ? v3 : -v3) + v_index - 1
					if backface
						# Vertex indices
						tri_data << v1
						tri_data << v3
						tri_data << v2
						# UV indices
						tri_data << v1
						tri_data << v3
						tri_data << v2
						# material index
						tri_data << mat_index
					else
						tri_data << v1
						tri_data << v2
						tri_data << v3
						# UV indices
						tri_data << v1
						tri_data << v2
						tri_data << v3
						# material index
						tri_data << mat_index
					end
				end
				v_index = v_index + facemesh_points.length
				#get_triangle_time += Time.now - t
				#puts "Getting triangles took #{Time.now - t} s"
				
				num_faces += 1
			end # End if entity == face
		end
		
		#puts "Added #{num_faces} faces."
		
		indigo_mesh.vert_pos_data = 	indigo_mesh.vert_pos_data.concat(position_data)
		indigo_mesh.vert_normal_data = 	indigo_mesh.vert_normal_data.concat(normal_data)
		indigo_mesh.uv_data = 			indigo_mesh.uv_data.concat(uv_data)
		indigo_mesh.triangle_data = 	indigo_mesh.triangle_data.concat(tri_data)
	end
	
	
	# Build and return an IndigoMesh object for the given entity list.
	# defn may be of type Sketchup::Model, or type Sketchup::ComponentDefinition
	def build_mesh(defn, ents)
		#puts "IndigoMeshBuilder::build_mesh()"
		#puts "defn: #{defn.inspect}"
		
		im = IndigoMesh.new(defn)
		
		
		position_data = []
		normal_data = []
		uv_data = []
		tri_data = []
		v_index = 0
		num_faces = 0
		
		add_material_time = 0.0
		make_face_mesh_time = 0.0
		get_vert_pos_time = 0.0
		get_vert_normal_time = 0.0
		get_uv_time = 0.0
		get_triangle_time = 0.0
		
		start_time = Time.now
		
		scale_factor = @scale
		
		ents.grep(Sketchup::Face) do |e|
			if e.visible? and e.layer.visible?
			
				#t = Time.now
			
				face = e
				mat = face.material
				back_mat = face.back_material
				backface = false
				if (mat==nil) and (back_mat!=nil)
					backface = true
					mat = back_mat
				end
		
				
				mat_index = im.add_material(mat)
				#add_material_time += Time.now - t
				
				#t = Time.now
				if backface
					facemesh = face.mesh(7)
				else
					facemesh = face.mesh(5)
				end
				#make_face_mesh_time += Time.now - t
				

				frontface = !backface

				
				# Get vertex positions
				#t = Time.now
				facemesh_points = facemesh.points
				facemesh_points.each do |point| 
					position_data << point.x * scale_factor
					position_data << point.y * scale_factor
					position_data << point.z * scale_factor
				end
				#get_vert_pos_time += Time.now - t
				#puts "Getting vertex positions took #{Time.now - t} s"
				
				# Get vertex normals
				#t = Time.now
				num_points = facemesh_points.length
				if frontface
					for i in 1..num_points
						n = facemesh.normal_at(i)
						normal_data << n.x
						normal_data << n.y
						normal_data << n.z
					end
				else
					for i in 1..num_points
						n = facemesh.normal_at(i)
						normal_data << -n.x
						normal_data << -n.y
						normal_data << -n.z
					end
				end
				#get_vert_normal_time += Time.now - t
				#puts "Getting vertex normals took #{Time.now - t} s"
				
				# Get vertex UVs
				#t = Time.now
				for i in 1..num_points
					uv = facemesh.uv_at(i, frontface)  # NOTE: can maybe use .uvs instead of .uv_at?
					#puts "uv: #{uv.to_a.inspect}"
					uv_data << uv.x / uv.z
					uv_data << uv.y / uv.z
				end
				#get_uv_time += Time.now - t
				#puts "Getting vertex UVs took #{Time.now - t} s"
				
				
				# Get triangles
				#t = Time.now
				polygons = facemesh.polygons
				for poly in polygons
					v1 = poly[0]
					v2 = poly[1]
					v3 = poly[2]
					# Convert to positive values, then convert to zero-based indices.
					v1 = (v1 >= 0 ? v1 : -v1) + v_index - 1
					v2 = (v2 >= 0 ? v2 : -v2) + v_index - 1
					v3 = (v3 >= 0 ? v3 : -v3) + v_index - 1
					if backface
						# Vertex indices
						tri_data << v1
						tri_data << v3
						tri_data << v2
						# UV indices
						tri_data << v1
						tri_data << v3
						tri_data << v2
						# material index
						tri_data << mat_index
					else
						tri_data << v1
						tri_data << v2
						tri_data << v3
						# UV indices
						tri_data << v1
						tri_data << v2
						tri_data << v3
						# material index
						tri_data << mat_index
					end
				end
				v_index = v_index + facemesh_points.length
				#get_triangle_time += Time.now - t
				#puts "Getting triangles took #{Time.now - t} s"
				
				num_faces += 1
			end # End if entity == face
		end
		
		total_time = Time.now - start_time
		
		if false
			puts ""
			puts "num_faces:            #{num_faces}"
			puts "add_material_time:    #{"%.4f" % add_material_time}"
			puts "make_face_mesh_time:  #{"%.4f" % make_face_mesh_time}"
			puts "get_vert_pos_time:    #{"%.4f" % get_vert_pos_time}"
			puts "get_vert_normal_time: #{"%.4f" % get_vert_normal_time}"
			puts "get_uv_time:          #{"%.4f" % get_uv_time}"
			puts "get_triangle_time:    #{"%.4f" % get_triangle_time}"
			puts "total_time:           #{"%.4f" % total_time}"
		end

		
		
		#puts "mat_to_index_map: #{mat_to_index_map.inspect}"
		
		im.vert_pos_data = position_data
		im.vert_normal_data = normal_data
		im.uv_data = uv_data
		im.triangle_data = tri_data
	
	
=begin		
		
		faces = {} # Map from Sketchup::Material to Array of (Sketchup::Face, bool) pairs.
		points = []
		normals = []
		uvs = []
		uv = []
		uv0 = []
		uv1 = []
		uv2 = []
		uv3 = []
		tris = []
		has_uvs = false
		has_uv0 = false


		face_triangles = []
		uvset0 = {}
		uvset1 = {}
		uvset2 = {}
		uvset3 = {}
		v_index = 0
		mat_index = 0

		#im = IndigoMesh.new(defn)
		
		active_uv_set = im.uv_set
		has_uv0 = im.has_uvset(0)
		has_uv1 = im.has_uvset(1)
		has_uv2 = im.has_uvset(2)
		has_uv3 = im.has_uvset(3)

		# Get a list of faces for each material
		for e in ents
			if (e.class==Sketchup::Face) and (e.visible?) and (e.layer.visible?)
				@faces = @faces + 1
				mat = e.material
				back_mat = e.back_material
				if (mat==nil) and (back_mat!=nil)
					backface = true
					mat = back_mat
				else
					backface = false
				end
				faces[mat] = [] if not faces[mat]
				faces[mat] << [e,backface]
			end
		end
		
		faces.each {|mat,face_arr|
			if(mat && !scene_materials.has_key?(mat))
				# puts "material found on face (#{mat.inspect}, name=#{mat.name.to_s}) was not found in scene materials!"
				next
			end
				
			mat_name = mat ? mat.name : "Default"
			
			face_triangles = []
			face_arr.each { |face, backface|

				e = face  #the face entity
				if backface
					facemesh=e.mesh(7)
				else
					facemesh = e.mesh(5)
				end
				#check the uvsets of the face
				#might be room for further optimization directly below
				if backface
					uvset0[facemesh]=e.back_material
				else
					uvset0[facemesh]=mat  #get the material actually applied to this face to determine if it contains sketchup default uvs or face-level uvs
				end
							
				facemesh_points=facemesh.points
				facemesh_points.each {|e| points<<e}
					
				#get the vertex normals
				for p in (1..facemesh_points.length)
					if backface
						normals.push(facemesh.normal_at(p).reverse)
					else
						normals.push(facemesh.normal_at(p))
					end
				end
				
				facemesh_uvs=[]
				for p in (1..facemesh_points.length)
					if backface
						facemesh_uvs.push(facemesh.uv_at(p,false))
					else
						facemesh_uvs.push(facemesh.uv_at(p,1))
					end
				end
				
				#if uvset0[facemesh]==nil #no material applied to face so apply the default SketchUp uvs
				#	facemesh_uvs=facemesh_uvs.collect {|p| [p.x,p.y/te,p.z]}
				#end
				facemesh_uvs.each {|e| uv<<e.to_a}
				#p uv
				#now the stored uvs sets
				#check if this mesh uses multiple uvsets and that the current face has the corresponding stored uvset
				#uv1=uv.dup
				#uv2=uv.dup
				#uv3=uv.dup
				
				#build any other stored uv sets
				if has_uv0
					#p "has uv0"
					#p facemesh_uvs
					uvs=face.get_attribute 'uvs','uv0',facemesh_uvs
					uvs.each {|e| uv0<<e.to_a}
					#p uv0
				end
				if has_uv1
					#p "has uv1"
					uvs=face.get_attribute 'uvs','uv1',facemesh_uvs
					uvs.each {|e| uv1<<e.to_a}
				end
				if has_uv2
					#p "has uv2"
					uvs=face.get_attribute 'uvs','uv2',facemesh_uvs
					uvs.each {|e| uv2<<e.to_a}
				end
				if has_uv3
					#p "has uv3"
					uvs=face.get_attribute 'uvs','uv3',facemesh_uvs
					uvs.each {|e| uv3<<e.to_a}
				end
				
				#note that the active uv set is the only set displayed by SketchUp and accessible directly by the API
				#p "active uv set #{active_uv_set}"
				
				#now the triangles
				polygons=facemesh.polygons
				for poly in polygons
					v1 = (poly[0]>=0?poly[0]:-poly[0])+v_index
					v2 = (poly[1]>=0?poly[1]:-poly[1])+v_index
					v3 = (poly[2]>=0?poly[2]:-poly[2])+v_index
					if backface
						face_triangles.push([v1-1,v3-1,v2-1])
					else
						face_triangles.push([v1-1,v2-1,v3-1])
					end
				end
				v_index=v_index+facemesh_points.length
			}

			im.add_triangles(face_triangles,mat)
		}

		case active_uv_set
			when 0
				uv0=uv
			when 1
				uv1=uv
			when 2
				uv2=uv
			when 3
				uv3=uv
		end

		im.add_points(points, @scale)  #takes an array of points
		im.add_normals(normals)  #takes an array of normals
		im.add_uvs(uv0, uv1, uv2, uv3)
=end

		return im # the IndigoMesh object
	end

	
	def get_linearised_mesh_for_definition(defn, overall_transformation)
	
		#puts "get_linearised_mesh_for_definition(), defn=#{defn.name}"
	
		info_list = get_linearised_definition_info_for_def(defn)
		
		#puts "info_list: #{info_list.inspect}"
	
		indigo_mesh = IndigoMesh.new(defn)
		indigo_mesh.flattened = true
		
		# Add top-level faces for the definition
		add_faces_to_mesh(defn.entities, indigo_mesh, overall_transformation, nil)
		
		info_list.each do |def_info|
			trans = def_info[0]
			mat = def_info[1]
			thedef = def_info[2]
			inst = def_info[3]
			
			if inst.class == Sketchup::Image
				#puts "found image: #{inst.path}"
				add_image_to_mesh(inst, indigo_mesh, overall_transformation * trans, mat)
			else
				#puts "Adding faces from defn #{thedef.name}..."
				add_faces_to_mesh(thedef.entities, indigo_mesh, overall_transformation * trans, mat)
			end
		end
		
		return indigo_mesh
	end
	
	
	
	
	
	
	def build_linearised_definition_info_for_instance(comp_instance, def_info_list)
		return if !comp_instance.visible? || !comp_instance.layer.visible?
		
		definition = comp_instance.definition # Get the component definition this comp_instance refers to:
		component_material = comp_instance.material # The component instance may have a material associated with it.
		
		res = definition.name.rindex("_dummy")
		if res != nil && res == (definition.name.length - 6) # if definition.name ends with '_dummy'.  (String.ends_width? is not available in Ruby 1.8.6)
			definition = get_dummy_defn(definition)
		end

		def_info_list << [comp_instance.transformation, component_material, definition, comp_instance]
		
		append_linearised_definition_info_for_def(definition, comp_instance.transformation, component_material, def_info_list) # Recurse
	end
	
	
	def build_linearised_definition_info_for_group(group, def_info_list)
		#puts "get_linearised_definition_info_for_group"
		return if !group.visible? || !group.layer.visible?
		
		definition = group.entities.parent
		component_material = group.material # The group may have a material associated with it.
		
		def_info_list << [group.transformation, component_material, definition, nil]
		
		append_linearised_definition_info_for_def(definition, group.transformation, component_material, def_info_list) # Recurse
	end
	
	
	# Gets or builds as necessary.
	def get_linearised_definition_info_for_def(definition)
		# puts "get_linearised_definition_info_for_def(), definition: #{definition.name} (#{definition.inspect})"
		subtree_def_info_list = @definition_info[definition]
		#puts "	found." if def_info_list
		if !subtree_def_info_list
			#puts "	not found!!, building #{definition.name}..."
			subtree_def_info_list = []
			build_linearised_definition_info_for_def(definition, subtree_def_info_list)
			@definition_info[definition] = subtree_def_info_list
		end
		return subtree_def_info_list
	end
	
	
	def append_linearised_definition_info_for_def(definition, overall_transform, component_material, def_info_list)
	
		subtree_def_info_list = get_linearised_definition_info_for_def(definition)
		subtree_def_info_list.each do |def_info|
			trans = def_info[0]
			mat = def_info[1]
			thedef = def_info[2]
			inst = def_info[3]
			
			def_info_list << [overall_transform * trans, mat ? mat : component_material, thedef, inst]
		end
	end
	
	
	def build_linearised_definition_info_for_image(image, def_info_list)
		# NOTE: Since Image.transformation is not defined in SketchUp < 2014, don't use transform now, just use identity.
		def_info_list << [Geom::Transformation.new, nil, nil, image]
	end
	
	
	def build_linearised_definition_info_for_def(definition, def_info_list)
		#puts "get_linearised_definition_info_for_def, num entities: #{definition.entities.length}"
		entities = definition.entities
		entities.grep(Sketchup::ComponentInstance) do |instance|
			build_linearised_definition_info_for_instance(instance, def_info_list) 
		end
		entities.grep(Sketchup::Group) do |group|
			build_linearised_definition_info_for_group(group, def_info_list)
		end
		entities.grep(Sketchup::Image) do |image|
			build_linearised_definition_info_for_image(image, def_info_list)
		end
	end
	
	
	# comp_instance - has type SketchUp::ComponentInstance
	def process_component_instance(comp_instance)
		#puts "IndigoMeshBuilder::process_component_instance()"
		
		return if !comp_instance.visible? || !comp_instance.layer.visible?

		definition = comp_instance.definition # Get the component definition this comp_instance refers to:
		component_material = comp_instance.material
		overall_transform = comp_instance.transformation
		
		# Handle special cases for sphere and _dummy (proxy component) export
		#if definition.name == "IndigoSphere"
		#	inst = IndigoInstance.new
		#	inst.transformation = overall_transform
		#	inst.su_instance = comp_instance
		#	inst.material = component_material
		#	@@spheres.push(inst)
		#	return
		#end
		
		res = definition.name.rindex("_dummy")
		if res != nil && res == (definition.name.length - 6) # if definition.name ends with '_dummy'.  (String.ends_width? is not available in Ruby 1.8.6)
			definition = get_dummy_defn(definition) # Replace the definition with the proxied definition.
		end
		
		# Build the IndigoInstance object
		indigo_inst = IndigoInstance.new
		indigo_inst.transformation = overall_transform
		indigo_inst.definition = definition
		indigo_inst.su_instance = comp_instance
		indigo_inst.material = component_material
		
		@@components_to_export.insert(definition) # Add to set of used definitions.
		@@instances << indigo_inst # Add to instance list

		
		# Now process each entity from the component definition
		def_info_list = get_linearised_definition_info_for_def(definition)
		
		def_info_list.each do |def_info|
			trans = def_info[0]
			mat = def_info[1]
			thedef = def_info[2]
			inst = def_info[3]
			
			if inst.class == Sketchup::Image
				@@images.push([inst, overall_transform * trans])
			else
				# Build the IndigoInstance object
				indigo_inst = IndigoInstance.new
				indigo_inst.transformation = overall_transform * trans
				indigo_inst.definition = thedef
				indigo_inst.su_instance = inst
				indigo_inst.material = mat ? mat : component_material
				
				@@components_to_export.insert(thedef) # Add to set of used definitions.
				@@instances << indigo_inst # Add to instance list
			end
		end
		
		#@num_components_processed += 1
	end
	
	
	# group - has type SketchUp::Group
	def process_group(group)
		#puts "IndigoMeshBuilder::process_group()"
		#puts "group.entities: #{group.entities.inspect}"
		#puts "group.entities.parent: #{group.entities.parent.inspect}"
		
		return if !group.visible? || !group.layer.visible?
		
		# NOTE that groups are internally implemented in SketchUp as Components.
		# We can use group.entities.parent to get the ComponentDefinition for the group.
		# This also seems to be the only way to get a GUID for the group for SketchUp < 2014.
		definition = group.entities.parent
		overall_transform = group.transformation
		group_material = group.material
		

		# Build the IndigoInstance object
		indigo_inst = IndigoInstance.new
		indigo_inst.transformation = overall_transform
		#indigo_inst.group = group
		indigo_inst.definition = definition
		indigo_inst.material = group_material
		
		# Add to instance list
		@@components_to_export.insert(definition)
		@@instances << indigo_inst
		

		# Now process each entity from the group
		def_info_list = get_linearised_definition_info_for_def(definition)
		def_info_list.each do |def_info|
			trans = def_info[0]
			mat = def_info[1]
			thedef = def_info[2]
			inst = def_info[3]
			
			if inst.class == Sketchup::Image
				@@images.push([inst, overall_transform * trans])
			else
				# Build the IndigoInstance object
				indigo_inst = IndigoInstance.new
				indigo_inst.transformation = overall_transform * trans
				indigo_inst.definition = thedef
				indigo_inst.su_instance = inst
				indigo_inst.material = mat ? mat : group_material
				
				@@components_to_export.insert(thedef) # Add to set of used definitions.
				@@instances << indigo_inst # Add to instance list
			end
		end
	end
	
	
	# image - has type SketchUp::Image
	def process_image(image)
		#puts "IndigoMeshBuilder::process_image()" if SkIndigo.verbose?
		
		# NOTE: Since Image.transformation is not defined in SketchUp < 2014, don't concatenate now.
		#overall_transform = transform * image.transformation
		overall_transform = Geom::Transformation.new
		
		@@images.push([image, overall_transform])
	end
	
	
	# Process an entities list
	# entities - an object of type Sketchup::Entities (http://www.sketchup.com/intl/en/developer/docs/ourdoc/entities)
	def process_entity_list(entities)
	
		#puts "IndigoMeshBuilder::process_entity_list()  (num ents: #{entities.length})"
		
		# Use the grep method to iterate over the entities, which is faster than interating over each entity in the entity list.
		# This is because a lot of the entities will be edges and faces, which we don't want to process here.
		entities.grep(Sketchup::ComponentInstance) do |instance|
			process_component_instance(instance) 
		end
		entities.grep(Sketchup::Group) do |group|
			process_group(group)
		end
		entities.grep(Sketchup::Image) do |image|
			process_image(image)
		end
	end
	
	
	def process_entities(entities)
	
		# Hash from Sketchup::ComponentDefinition to array of [transformation, component material, definition, instance], of all the instances in that definition.
		@definition_info = {}
		
		# Do a recursive traversal over the entity tree.
		process_entity_list(entities)
		
		# Make an instance for the top-level faces.
		inst = IndigoInstance.new
		inst.transformation = Geom::Transformation.new
		@@instances << inst # Add to instance list
		
		@@top_level_entities_to_export = entities # Remember these entities, to convert to a mesh later.
	end
	
	
	def get_dummy_defn(definition)
		name = definition.name
		
		dummy_def_name = name[0, name.length - 6] # remove "_dummy" suffix
		
		dummy_def = Sketchup.active_model.definitions[dummy_def_name]
		
		return dummy_def ? dummy_def : definition
	end
end  #class end
